package core;

import java.io.FileInputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.math.BigInteger;
import java.net.ServerSocket;
import java.net.Socket;
import java.security.Key;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.SecureRandom;
import java.security.Security;
import java.security.cert.Certificate;
import java.util.Stack;
import java.util.Vector;

import javax.crypto.SecretKey;
import javax.crypto.SecretKeyFactory;
import javax.crypto.spec.DESKeySpec;

import org.bouncycastle.jce.provider.BouncyCastleProvider;
//import cryptix.jce.provider.CryptixCrypto;

/*
 * Created on May 11, 2003
 */

/**
 * Any extension of this abstract class represents a security protocol from the viewpoint of one of its agents.
 * For example, to implement a Protocol with n agents, we need n classes extending Protocol. 
 * @author Xavier Didelot
 */
public abstract class Protocol extends Thread {

	/**
	 * The SecretKeyStore in which all secret/shared keys are stored
	 */
	private SecretKeyStore secretKeyStore;

	/**
	 * The KeyStore in which all public/private keys are stored
	 */
	private KeyStore myKeyStore;
	/**
	 * This is the password to use to access the KeyStore
	 */
	private char[] password;
	/**
	 * This vector contains the variables that this Protocol knows (of type Nonce, Integer and Agent)
	 */
	protected Vector variables;

	/**
	 * This stack is used to store the different fields of a given message
	 */
	private Stack parts;
	/**
	 * This stack is used to restore a previous state during the protocol run
	 */
	private Stack stackM0;
	/**
	 * This message is used to manipulate the fields of a Message
	 */
	protected Message m0;
	
	protected ObjectOutputStream oos;
	
	protected static String protoTag="";

	private static Class messageC;
	private static Class messageKeyC;
	private static Class nonceC;

	protected Message message() {
		Class[] a={};
		try {
			return (Message)messageC.getDeclaredConstructor(a).newInstance(a);
		} catch (Exception e) {e.printStackTrace();return null;}
	}
	
	protected Message message(String s) {
		try {
			Message m=message();
			m.setString(s);
			return m;
		} catch (Exception e) {e.printStackTrace();return null;}
	}
	
	protected Message message(Message m1, Message m2) {
		try {
			Class [] a={Class.forName("core.Message"),Class.forName("core.Message")};
			Object [] b={m1,m2};
			return (Message)messageC.getDeclaredConstructor(a).newInstance(b);
		} catch (Exception e) {e.printStackTrace();return null;}
	}
	
	protected Message message(Agent agent) {
		try {
			Class [] a={Class.forName("core.Agent")};
			Object [] b={agent};
			return (Message)messageC.getDeclaredConstructor(a).newInstance(b);
		} catch (Exception e) {e.printStackTrace();return null;}
	}
	
	protected Message message(INonce n) {
		try {
			Class [] a={Class.forName("core.INonce")};
			Object [] b={n};
			return (Message)messageC.getDeclaredConstructor(a).newInstance(b);
		} catch (Exception e) {e.printStackTrace();return null;}
	}
	
	protected INonce nonce() {
		try {
			Class [] a={};
			return (INonce)nonceC.getDeclaredConstructor(a).newInstance(a);
		} catch (Exception e) {e.printStackTrace();return null;}
	}
	
	protected INonce nonce(Integer i) {
		try {
			Class [] a={Class.forName("java.lang.Integer")};
			Object [] b = {i};
			return (INonce)nonceC.getDeclaredConstructor(a).newInstance(b);
		} catch (Exception e) {e.printStackTrace();return null;}
	}
	
	protected INonce nonce(String s) {
		try {
			Class [] a={Class.forName("java.lang.String")};
			Object [] b={s};
			return (INonce)nonceC.getDeclaredConstructor(a).newInstance(b);
		} catch (Exception e) {e.printStackTrace();return null;}
	}
	
	protected IMessageKey messageKey() {
		try {
			Class [] a={};
			return (IMessageKey)messageKeyC.getDeclaredConstructor(a).newInstance(a);
		} catch (Exception e) {e.printStackTrace();return null;}
	}
	
	protected IMessageKey messageKey(String algo) {
		try {
			Class [] a={Class.forName("java.lang.String")};
			Object [] b={algo};
			return (IMessageKey)messageKeyC.getDeclaredConstructor(a).newInstance(b);
		} catch (Exception e) {e.printStackTrace();return null;}
	}
	
	protected IMessageKey messageKey(Key key) {
		try {
			Class [] a={Class.forName("java.security.Key")};
			Object [] b={key};
			return (IMessageKey)messageKeyC.getDeclaredConstructor(a).newInstance(b);
		} catch (Exception e) {e.printStackTrace();return null;}
	}
	
	protected IMessageKey messageKey(Message msg) {
		try {
			Class [] a={Class.forName("core.Message")};
			Object [] b={msg};
			return (IMessageKey)messageKeyC.getDeclaredConstructor(a).newInstance(b);
		} catch (Exception e) {e.printStackTrace();return null;}
	}		

	protected void checkTag(Message mm, String s) throws Exception {
		MessageT m=(MessageT)mm;
		if (m.getTag().getCode().compareTo(s)!=0) 
			throw new Exception("Invalid tag: "+s+" expected, "+m.getTag().getCode()+" received");
		println("Tag check successful for "+s);
		m.removeTags();
		//m.setString(m.getText());
		//System.out.println(m.getString());
	}



	/**
	 * This method saves the current value of Message m0
	 */
	protected void storeM0() {
		stackM0.push(m0);
		m0 = message();
	}

	/**
	 * This method restore the last value of Message m0
	 * @return the last value of Message m0
	 */
	protected Message restoreM0() {
		return (Message) stackM0.pop();
	}

	/**
	 * This method returns the next field to be treated
	 * @return next field to be treated
	 */
	protected Message getPart() {
		return (Message) parts.pop();
	}

	/**
	 * This method adds an agent to the list of known agents
	 * @param agent agent to add to the list of known agents
	 */
	protected void addAgent(Agent agent) {
		variables.addElement(agent);
	}

	/**
	 * This method adds a nonce to the list of known nonces
	 * @param nonce nonce to add to the list of known nonces
	 */
	protected void addNonce(INonce nonce) {
		variables.addElement(nonce);
	}

	/**
	 * Creates a new instance of Protocol
	 */
	public Protocol(String m, String mk, String n) {
		try {
			messageC=Class.forName(m);
			messageKeyC=Class.forName(mk);
			nonceC=Class.forName(n);
		} catch (Exception e) {e.printStackTrace();}
		//Security.addProvider(new CryptixCrypto());
		//Security.addProvider(new ForgeProvider());
		Security.addProvider(new BouncyCastleProvider());
		//Security.addProvider(new IAIK());
		Message m0 = message();
		parts = new Stack();
		stackM0 = new Stack();
		variables = new Vector();
	}

	/**
	 * this method sets the SecretKeyStore that has to be used
	 * @param file the SecretKeyStore that has to be used
	 */
	protected void setSKS(String file) {
		if (file != null)
			secretKeyStore = new SecretKeyStore(file);
		else
			secretKeyStore = null;
	}
	
	/**
	 * this method sets the KeyStore that has to be used
	 * @param fileKeyStore the KeyStore that has to be used
	 * @param password password to use to access this KeyStore
	 */
	protected void setKS(String fileKeyStore, String password) {
		this.password = password.toCharArray();
		myKeyStore = null;
		try {
			myKeyStore = KeyStore.getInstance("JKS");
			myKeyStore.load(new FileInputStream(fileKeyStore), this.password);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * This method adds the given Messages into the list of Messages to be treated
	 * @param array array of messages that will have to be treated
	 * @param size expected size of the array
	 * @throws Exception if the Message does not have the expected number of fields
	 */
	protected void addParts(IMessage[] array, int size) throws Exception {
		if (array.length != size)
			throw new Exception("Message does not have the expected form"+array.length+size);
		int i;
		for (i = array.length - 1; i >= 0; i--)
			parts.push(array[i]);
	}

	/**
	 * This method returns the Private Key of a given agent by looking after it in the KeyStore
	 * @param agent the agent whose key we are looking for
	 * @return the private key of the given agent
	 */
	public PrivateKey getPrivateKey(String agent) {
		Key key = null;
		try {
			key = myKeyStore.getKey(agent, password);
		} catch (Exception e) {
			e.printStackTrace();
		}
		return (PrivateKey) key;
	}

	/**
	* This method returns the Public Key of a given agent by looking after it in the KeyStore
	* @param agent the agent whose key we are looking for
	* @return the public key of the given agent
	*/
	public PublicKey getPublicKey(String agent) {
		Certificate certificate = null;
		try {
			certificate = myKeyStore.getCertificate(agent);
		} catch (KeyStoreException e) {
			e.printStackTrace();
		}
		PublicKey k = null;
		try {
			k = certificate.getPublicKey();
		} catch (NullPointerException e) {
			System.out.println(agent);
			e.printStackTrace();
		}
		return k;
	}
	
	/**
	* This method returns the Secret Key of a given agent by looking after it in the SecretKeyStore
	* @param agent the agent whose key we are looking for
	* @return the secret key of the given agent
	*/
	public Key getSecretKey(String id)
		throws IOException, ClassNotFoundException {
		return secretKeyStore.getKey(id);
	}

	/**
	 * get Agent from id 
	 * @param id id of the agent to find
	 * @return the Agent corresponding to this id
	 */
	protected Agent agentFromId(String id) {
		int i = 0;
		while (i < variables.size()) {
			try {if (id.compareTo(((Agent) variables.elementAt(i)).getId()) == 0) break;}
				catch (ClassCastException e) {}
			i++;}
		if (i >= variables.size()) {
			Agent agent = new Agent();
			agent.setId(id);
			addAgent(agent);
			return agentFromId(id);
		}
		return ((Agent) variables.elementAt(i));
	}

	/**
	 * This method returns the Nonce whose id is given
	 * @param id id of the Nonce we are looking for
	 * @return the Nonce whose id is given
	 */
	public INonce nonceFromId(String id) {
		int i = 0;
		while (i < variables.size()) {
			try {if (id.compareTo(((INonce) variables.elementAt(i)).getId()) == 0) break;}
				catch (ClassCastException e) {}
			i++;}
		if (i >= variables.size()) {
			INonce nonce =nonce();
			nonce.setId(id);
			addNonce(nonce);
			return nonceFromId(id);
		}
		return ((INonce) variables.elementAt(i));
	}

	/**
	 * This method returns the name of the agent whose id is given
	 * @param id id of the agent whose name we are looking for
	 * @return the name of the agent whose id is given
	 */
	protected String nameFromId(String id) throws Exception {
		Agent agent = agentFromId(id);
		return agent.getName();
	}

	/**
	 * This method returns the IP of the agent whose id is given
	 * @param id id of the agent whose name we are looking for
	 * @return the IP of the agent whose id is given
	 */
	protected String IPFromId(String id) throws Exception {
		Agent agent = agentFromId(id);
		if (agent.getIp()==null) agent.setIP(agent.getSocket().getInetAddress().getHostAddress());
		return agent.getIp();
	}

	/**
	 * Waits for a message from the given agent and returns it. 
	 * @param id id of the agent from which to get a message
	 * @return the next message received
	 */
	public String get(String id,int port) {
		Agent agent = null;
		try {
			agent = agentFromId(id);
		} catch (Exception e) {
			e.printStackTrace();
		}
		return agent.get(port);
	}
	
	public String get(String id) {
		return get(id,4000);
	}

	/** 
	 * Waits for a  connection to be established. The id assigned to this agent is the arguement id
	 * @param id id to assign to the new host
	 * @param port port on which we are waiting for a connection
	 */
	public void newAgent(String id, int port) {
		try {
			ServerSocket ss = new ServerSocket(port);
			Socket clientSocket = null;
			clientSocket = ss.accept();
			Agent agent = new Agent();
			agent.setId(id);
			agent.setSocket(clientSocket);
			addAgent(agent);
			ss.close();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * Sends a message to another agent
	 * @param name name of the agent we are sending a message to
	 * @param what the message to send
	 */
	public void put(String id, String what) {
		put(id,what,4000);
	}

	/**
	 * Sends a message to another agent using the specified port
	 * @param name name of the agent we are sending a message to
	 * @param what the message to send
	 * @param port port to use to send the message
	 */
	public void put(String id, String what,int port) {
		Agent agent = null;
		try {
			agent = agentFromId(id);
		} catch (Exception e) {
			e.printStackTrace();
		}
		agent.put(what,port);
	}

	/**
	 * This method sets the name of a given Agent
	 * @param name name to give to the Agent
	 * @param id id of the Agent
	 * @throws Exception if the name of this Agent has already been set to something else
	 */
	public void setNameToId(String name, String id) throws Exception {
		Agent agent;
		agent = agentFromId(id);
		if (agent.getName().compareTo("") == 0)
			agent.setName(name);
		else if (agent.getName().compareTo(name) != 0)
			throw new Exception("Incorrect name received for Agent " + id+" expected="+agent.getName()+" received="+name);
	}

	/**
	 * This method sets the IP of a given Agent
	 * @param ip IP to give to the Agent
	 * @param id id of the Agent
	 */
	public void setIPToId(String ip, String id) {
		Agent agent;
		try {
			agent = agentFromId(id);
		} catch (Exception e) {
			agent = new Agent();
			agent.setId(id);
			addAgent(agent);
		}
		try {
			if (ip.startsWith("D,")) agent.setIP(ip.substring(2)); else agent.setIP(ip);
		} catch (Exception e1) {
			e1.printStackTrace();
		}
	}

	/**
	 * This method sets the value of a given Nonce
	 * @param value value to give to the Nonce
	 * @param id id of the Nonce
	 * @throws Exception if the value of this Nonce has already been set to something else
	 */
	public void setValueToNonce(String value, String id) throws Exception {
		INonce nonce;
		nonce = nonceFromId(id);
		if (nonce.getString().compareTo("") == 0)
			nonce.setString(value);
		else if (nonce.getString().compareTo(value) != 0)
			throw new Exception(
				"Incorrect value received for Nonce "
					+ id
					+ " "
					+ nonce.getString()
					+ " "
					+ value);
	}

	/**
	 * This method creates a new Nonce
	 * @param id id of the new Nonce to create
	 */
	public void createNewNonce(String id) {
		INonce n = nonce(new Integer(0));
		n.setId(id);
		variables.addElement(n);
	}

	/**
	 * This method returns the value of a given Nonce
	 * @param id id of the Nonce whose value we are looking for
	 * @return the value of the given Nonce
	 */
	public String valueFromNonce(String id) {
		INonce nonce;
		nonce = nonceFromId(id);
		return nonce.getString();
	}
	
	/**
	 * This method has to be implemented by classes extending the Protocol class.
	 * It represents the actions to perform when the protocol is started.
	 */
	public abstract void run();

	/**
	 * This method has to be implemented by classes extending the Protocol class.
	 * It represents the actions to perform to get the arguments of the protocol.
	 */
	public abstract String[] getArgs(String[] args);
		
	/**
	 * This method returns the Vector of Agents
	 * @return the vector of Agents
	 
	public Vector getAgents() {
		return agents;
	}*/

	public void matchXor(Message m4, String string, String string2) throws Exception {
		INonce n=nonceFromId(string);
		INonce n2=nonceFromId(string2);
		INonce ntest=nonce();
		INonce na,nb;
		if (n.getString().compareTo(ntest.getString())==0 && n2.getString().compareTo(ntest.getString())!=0)
			{na=n; nb=n2;} else
		if (n.getString().compareTo(ntest.getString())!=0 && n2.getString().compareTo(ntest.getString())==0)
		{na=n2; nb=n;} else throw new Exception("matchXor impossible "+n.getString()+" "+n2.getString());
		IMessage m=(IMessage)m4.clone();
		m.unxor(nb);
		na.setString(m.getString());
	}

	protected synchronized void println(String str)
	{
		String name=getClass().getName();
		name=name.substring(name.lastIndexOf(".")+1);
		System.out.println(name+": "+str);
	}
	
	protected static BigInteger freshBI() {
		SecureRandom sr=new SecureRandom();
		byte[] pseudoRandom = new byte[10];
		sr.nextBytes(pseudoRandom);
		return new BigInteger(pseudoRandom).abs();		
	}

	protected static SecretKey generateKey(byte[] password) throws Exception
	{
	  DESKeySpec ks = new DESKeySpec(password);
	  SecretKeyFactory skf = SecretKeyFactory.getInstance("DES");
	  SecretKey key = skf.generateSecret(ks);
	  return key;
	}

	protected void  writeObject(Object obj) throws IOException {
		synchronized (oos) {
		oos.writeObject(obj);
		oos.notify();
		}
	}

}



